﻿// ----- Time-Based One-Time Password Example Application
//       Copyright (c) 2019 by Tim Patrick
//       Visit: www.owanipress.com/totp

using System;
using System.ComponentModel;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using QRCoder;

namespace ExampleTOTP
{
    public class MainWindowViewModel : INotifyPropertyChanged
    {
        private const string IssuerName = "Owani Press";
        private SimpleFileOtp CurrentUser = null;

        public event PropertyChangedEventHandler PropertyChanged;
        public void NotifyPropertyChanged(string propName)
        {
            this.PropertyChanged(this, new PropertyChangedEventArgs(propName));
        }

        // ----- Entry field for email address.
        private string EmailAddressInternal = "";
        public string EmailAddress
        {
            get
            {
                return this.EmailAddressInternal;
            }
            set
            {
                // ----- Any change forces a sign-out.
                this.EmailAddressInternal = value;
                this.SignOut();
                NotifyPropertyChanged(nameof(this.EmailAddress));
            }
        }

        // ----- Entry field for pass code.
        private string PassCodeInternal = "";
        public string PassCode
        {
            get
            {
                return this.PassCodeInternal;
            }
            set
            {
                this.PassCodeInternal = value;
                NotifyPropertyChanged(nameof(this.PassCode));
            }
        }

        // ----- Status fields.
        private ImageSource QRCodeInternal = null;
        public ImageSource QRCode
        {
            get
            {
                return this.QRCodeInternal;
            }
            private set
            {
                this.QRCodeInternal = value;
                NotifyPropertyChanged(nameof(this.QRCode));
            }
        }

        private bool QRCodeVisibleInternal = false;
        public bool QRCodeVisible
        {
            get
            {
                return this.QRCodeVisibleInternal;
            }
            private set
            {
                this.QRCodeVisibleInternal = value;
                NotifyPropertyChanged(nameof(this.QRCodeVisible));
            }
        }

        private bool SignedInInternal = false;
        public bool SignedIn
        {
            get
            {
                return this.SignedInInternal;
            }
            private set
            {
                this.SignedInInternal = value;
                NotifyPropertyChanged(nameof(this.SignedIn));
            }
        }

        public string ErrorMessage { get; private set; }

        private void SignOut()
        {
            // ----- Sign out, clearing all fields except the email address.
            if (this.CurrentUser != null)
            {
                this.CurrentUser = null;
                this.PassCode = "";
                this.QRCode = null;
                this.QRCodeVisible = false;
                this.SignedIn = false;
                this.ErrorMessage = null;
            }
        }

        public bool SignIn()
        {
            // ----- Access an account based on an email address.

            // ----- Initialize.
            this.SignOut();

            // ----- Check for invalid data.
            if (string.IsNullOrWhiteSpace(this.EmailAddress))
            {
                this.ErrorMessage = "Please provide an email address.";
                return false;
            }

            // ----- Access the account. This can thrown an invalid-email exception.
            try
            {
                this.CurrentUser = new SimpleFileOtp();
                this.CurrentUser.Issuer = IssuerName;
                this.CurrentUser.EmailAddress = this.EmailAddress.Trim();
            }
            catch (Exception ex)
            {
                this.ErrorMessage = ex.Message;
                this.CurrentUser = null;
                return false;
            }

            // ----- Valid user. Generate key if this is a first-time access.
            if (this.CurrentUser.HasSecretKey == false)
                if (this.CurrentUser.GenerateKey(true) == false)
                {
                    this.ErrorMessage = "Unable to create access key for two-factor authentication.";
                    this.CurrentUser = null;
                    return false;
                }

            // ----- Success.
            this.SignedIn = true;
            return true;
        }

        public bool DisplayQRCode()
        {
            // ----- Display a QR code for the active account.
            QRCode processQR;

            // ----- Perhaps the QR code is already visible.
            if (this.QRCodeVisible == true)
                return true;

            // ----- This only works if a user is active.
            this.ErrorMessage = null;
            if (this.SignedIn == false)
            {
                this.ErrorMessage = "Please sign in to display the QR code.";
                return false;
            }

            // ----- Create and display the QR code. Based on:
            //         https://github.com/codebude/QRCoder
            //       Then, convert to WPF image format.
            try
            {
                processQR = new QRCode(new QRCodeGenerator().CreateQrCode(
                    this.CurrentUser.LinkPath, QRCodeGenerator.ECCLevel.Q));
                this.QRCode = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
                    processQR.GetGraphic(5).GetHbitmap(), IntPtr.Zero, Int32Rect.Empty,
                    BitmapSizeOptions.FromEmptyOptions());
            }
            catch
            {
                this.ErrorMessage = "Unable to create QR Code for access link.";
                this.QRCode = null;
                return false;
            }

            // ----- Success.
            this.QRCodeVisible = true;
            return true;
        }

        public bool CheckPassCode()
        {
            // ----- Check an account-specific and time-specific pass code.

            // ----- This only works if a user is active.
            this.ErrorMessage = null;
            if (this.SignedIn == false)
            {
                this.ErrorMessage = "Please sign in to validate a pass code.";
                return false;
            }

            // ----- Check for missing data.
            if (string.IsNullOrWhiteSpace(this.PassCode))
            {
                this.ErrorMessage = "Please provide a pass code.";
                return false;
            }

            // ----- Check the pass code.
            if (this.CurrentUser.CheckCode(this.PassCode.Trim(), DateTime.Now) == false)
            {
                this.ErrorMessage = "Invalid pass code.";
                return false;
            }

            // ----- Success.
            return true;
        }
    }
}

