﻿// ----- Luigi
//       Web-based reporting system
//       Copyright (c) 2019 by Pagliacci Pizza

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using static ExampleTOTP.SupportMethods;

namespace ExampleTOTP
{
    public class SimpleFileOtp : TimeBasedPassword
    {
        public SimpleFileOtp()
        {
            // ----- Retrieve existing data, if needed.
            if (EmailSecretKeys.Count == 0)
                this.LoadKeyData();
        }

        private string EmailForStorage
        {
            get
            {
                // ----- Return a standardized version of the email for database storage.
                if (string.IsNullOrWhiteSpace(this.EmailAddress))
                    return "";
                else
                    return this.EmailAddress.Trim().ToLower();
            }
        }

        protected override void GetIDFromEmail()
        {
            // ----- This class doesn't use IDs, but that field can't remain null.
            this.UserID = new object();
        }

        public override bool HasSecretKey
        {
            get
            {
                // ----- Return True if the secret key exists. WARNING: Don't access
                //       the UseStatus member here as it will lead to an infinite loop.
                if (string.IsNullOrWhiteSpace(this.EmailAddress))
                    return false;
                else
                    return EmailSecretKeys.ContainsKey(this.EmailForStorage);
            }
        }

        protected override string RetrieveSecretKey()
        {
            // ----- Retrieve the current secret key for this user.

            // ----- Don't bother if no user has been indicated.
            if (this.UseStatus == TimeBasedPasswordStatus.NoEmailAddress)
                return null;

            // ----- Retrieve the key.
            if (EmailSecretKeys.ContainsKey(this.EmailForStorage))
                return EmailSecretKeys[this.EmailForStorage];
            else
                return null;
        }

        protected override bool StoreSecretKey(string secretKey)
        {
            // ----- Store a new secret key for this user.

            // ----- Don't bother if no user has been indicated.
            if (this.UseStatus == TimeBasedPasswordStatus.NoEmailAddress)
                return false;

            // ----- Update or remove this key?
            if (string.IsNullOrWhiteSpace(secretKey))
            {
                // ----- Remove the key, but keep the interval list just in case.
                //       It will be purged on save if irrelevant.
                if (EmailSecretKeys.ContainsKey(this.EmailForStorage))
                    EmailSecretKeys.Remove(this.EmailForStorage);
            }
            else if (EmailSecretKeys.ContainsKey(this.EmailForStorage))
            {
                // ----- Update the existing key.
                EmailSecretKeys[this.EmailForStorage] = secretKey;
            }
            else
            {
                // ----- Add a new key.
                EmailSecretKeys.Add(this.EmailForStorage, secretKey);
                if (EmailIntervals.ContainsKey(this.EmailForStorage) == false)
                    EmailIntervals.Add(this.EmailForStorage, new List<long>());
            }

            // ----- Save data immediately.
            this.SaveKeyData();
            return true;
        }

        protected override bool IntervalAvailable(long whichInterval)
        {
            // ----- Return False if the user already guessed the interval successfully.

            // ----- Don't bother if no user has been indicated.
            if (this.UseStatus == TimeBasedPasswordStatus.NoEmailAddress)
                return false;

            // ----- Look for the interval.
            if (EmailIntervals.ContainsKey(this.EmailForStorage))
                return !EmailIntervals[this.EmailForStorage].Contains(whichInterval);
            else
                return true;
        }

        protected override void DisableInterval(long whichInterval)
        {
            // ----- This interval has been used.

            // ----- Don't bother if no user has been indicated.
            if (this.UseStatus == TimeBasedPasswordStatus.NoEmailAddress)
                return;

            // ----- Look for the interval.
            if (EmailIntervals.ContainsKey(this.EmailForStorage) == false)
                EmailIntervals.Add(this.EmailForStorage, new List<long>());
            if (EmailIntervals[this.EmailForStorage].Contains(whichInterval) == false)
                EmailIntervals[this.EmailForStorage].Add(whichInterval);

            // ----- Save data immediately.
            this.SaveKeyData();
        }

        // ===== Insecure flat file database management =====
        //       Each line of the file has this format:
        //         name@exmaple.com,secret_key,interval,interval,interval...
        //       That is, an email address paired with a secret key and zero
        //       or more saved intervals.
        private const string DataFile = "OtpKeyData.txt";
        private static Dictionary<string, string> EmailSecretKeys = new Dictionary<string, string>();
        private static Dictionary<string, List<long>> EmailIntervals = new Dictionary<string, List<long>>();

        private void LoadKeyData()
        {
            // ----- Retrive all data from the file and store in memory.
            StreamReader sourceFile;
            string oneLine;
            string[] lineParts;
            string emailAddress;
            string secretKey;
            List<long> intervals;

            // ----- Reset the data.
            EmailSecretKeys.Clear();
            EmailIntervals.Clear();

            // ----- Try to access the file.
            try
            {
                if (File.Exists(DataFile) == false)
                    return;
                sourceFile = new StreamReader(DataFile);
            }
            catch
            {
                // ----- Some failure. Pretend the file doesn't exist.
                return;
            }

            // ----- Scan through the file.
            while (sourceFile.EndOfStream == false)
            {
                // ----- Read the line.
                try
                {
                    oneLine = sourceFile.ReadLine();
                }
                catch
                {
                    oneLine = "";
                }

                // ----- Get the parts. At the very least, two parts are needed: an
                //       email address and a secret key.
                lineParts = oneLine.Split(',');
                if (lineParts.Length < 2)
                    continue;

                // ----- Identify the parts.
                emailAddress = lineParts[0].Trim().ToLower();
                secretKey = lineParts[1].Trim();
                if ((emailAddress.Length == 0) | (secretKey.Length == 0))
                    continue;
                intervals = new List<long>();
                for (int position = 2; position < lineParts.Length; position++)
                    if (long.TryParse(lineParts[position], out long testInterval) == true)
                        if (intervals.Contains(testInterval) == false)
                            intervals.Add(testInterval);

                // ----- Store the parts.
                if (EmailSecretKeys.ContainsKey(emailAddress) == false)
                {
                    EmailSecretKeys.Add(emailAddress, secretKey);
                    EmailIntervals.Add(emailAddress, intervals);
                }
            }
            sourceFile.Close();
        }

        private void SaveKeyData()
        {
            // ----- Bundle up the data and save it to a file.
            StringBuilder content;
            List<long> cleanIntervals;

            // ----- Process each email address.
            content = new StringBuilder();
            foreach (string scanEmail in EmailSecretKeys.Keys)
            {
                // ----- Add the email and secret key.
                content.Append($"{scanEmail},{EmailSecretKeys[scanEmail]}");

                // ----- Add the intervals, if available.
                if (EmailIntervals.ContainsKey(scanEmail))
                {
                    cleanIntervals = PurgeOldIntervals(EmailIntervals[scanEmail]);
                    if ((cleanIntervals != null) && (cleanIntervals.Count > 0))
                        content.Append(",").Append(string.Join(",", cleanIntervals));
                }

                content.AppendLine();
            }

            // ----- Save the data to the file.
            try
            {
                File.WriteAllText(DataFile, content.ToString());
            }
            catch
            {
                // ----- Failure, but no logging in this sample.
            }
        }

        private List<long> PurgeOldIntervals(List<long> currentIntervals)
        {
            // ----- Remove old, and therefore inaccessible, intervals.
            List<long> cleanSet;
            long oldestToKeep;

            // ----- Check for missing data.
            if (currentIntervals == null)
                return new List<long>();
            else if (currentIntervals.Count == 0)
                return currentIntervals;

            // ----- Determine the lower boundary for retaining intervals.
            oldestToKeep = IntervalFromTime(DateTime.Now, this.IntervalSeconds) - this.IntervalDrift - 2;

            // ----- Build a new list, omitting the old intervals.
            cleanSet = new List<long>();
            foreach (long scanInterval in currentIntervals)
                if (scanInterval >= oldestToKeep)
                    cleanSet.Add(scanInterval);

            return cleanSet;
        }
    }
}

