﻿// ----- Luigi
//       Web-based reporting system
//       Copyright (c) 2019 by Pagliacci Pizza

using System;
using System.Text;
using System.Text.RegularExpressions;

namespace ExampleTOTP
{
    // ----- Support methods, not specific to the TOTP algorithm.
    public static class SupportMethods
    {
        public static string Encode32(string content)
        {
            // ----- Encode a text string as base-32, using a variation of RFC 3548.
            string allBits;
            string encoded;
            byte[] contentAsBytes;

            // ----- The IETF standard defines the following 32-element alphabet.
            char[] allowedSet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567".ToCharArray();

            // ----- Don't bother if there is no source content.
            if (string.IsNullOrEmpty(content))
                return "";

            // ----- Convert each incoming character to its binary equivalent.
            //       This code should use a StringBuilder for speed, but since
            //       this demonstration will only work with content strings up
            //       to 32 characters, we'll endure the consequences. Also, this
            //       code assumes that the original string is ASCII encoded.
            allBits = "";
            contentAsBytes = Encoding.ASCII.GetBytes(content);
            foreach (byte scanByte in contentAsBytes)
                allBits += Convert.ToString(scanByte, 2).PadLeft(8, '0');

            // ----- The binary version will be processed in 5-bit chunks.
            //       The standard requires that this input be in 40-bit
            //       chunks, but the standard used for the shared secret
            //       stipulates that this padding is not necessary. Add
            //       enough zeros to enable full 5-bit chunks.
            allBits += new string('0', (5 - (allBits.Length % 5)) % 5);

            // ----- Convert each 5-bit chunk back to an integer, and use
            //       it as an index into the set of allowed characters.
            encoded = "";
            for (int position = 0; position < allBits.Length; position += 5)
                encoded += allowedSet[Convert.ToInt32(allBits.Substring(position, 5), 2)];
            return encoded;
        }

        public static bool ConstantTimeCompare(string testValue, string realValue)
        {
            // ----- Compare two strings so that as consistent amount of elapsed time
            //       as possible is used for every comparison. This helps dissuade
            //       hacking behavior. Return True if the values are equal.
            uint mistakeMonitor;
            char realCharacter;

            // ----- If either value is null, it will mess up the comparison time, but
            //       that not this's method's fault.
            if ((testValue == null) | (realValue == null))
                return false;

            // ----- First, compare the lengths.
            mistakeMonitor = (uint)testValue.Length ^ (uint)realValue.Length;

            // ----- Next, compare every single character, up to the length of the
            //       test (user-supplied) value.
            for (int position = 0; position < testValue.Length; position++)
            {
                realCharacter = (position >= realValue.Length ? '\0' : realValue[position]);
                mistakeMonitor |= (uint)testValue[position] ^ (uint)realCharacter;
            }

            // ----- If any differences were found, they will appear as non-zero bits.
            return (mistakeMonitor == 0);
        }

        public static long IntervalFromTime(DateTime dateTime, int secondsPerInterval)
        {
            DateTime unixEpoch;
            TimeSpan elapsedTime;

            // ----- The UNIX epoch (T0) began at midnight on January 1, 1970.
            unixEpoch = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);

            // ----- Calculate number of intervals since T0. Calculation must be
            //       as a long integer to prevent failure on January 19, 2038.
            elapsedTime = dateTime.ToUniversalTime() - unixEpoch;
            return (long)Math.Floor(elapsedTime.TotalSeconds / secondsPerInterval);
        }

        public static bool ValidateEmailAddress(string testEmail)
        {
            // ----- Check for missing email.
            if (string.IsNullOrWhiteSpace(testEmail))
                return false;

            // ----- Make sure the address looks like an email address.
            const string emailPattern =
                @"^[A-Za-z0-9!#$%&'*+\/=?^_`{|}~-]+(?:\.[A-Za-z0-9!#$%&'*+\/=?^_`{|}~-]+)*@" +
                @"(?:[A-Za-z0-9](?:[A-Za-z0-9-]*[A-Za-z0-9])?\.)+[A-Za-z0-9](?:[A-Za-z0-9-]*[A-Za-z0-9])?$";
            try
            {
                return Regex.IsMatch(testEmail.Trim(), emailPattern);
            }
            catch
            {
                return false;
            }
        }
    }
}

