// ----- Luigi
//       Web-based reporting system
//       Copyright (c) 2019 by Pagliacci Pizza

using System;
using System.Security.Cryptography;
using System.Text;
using static System.Uri;
using static ExampleTOTP.SupportMethods;

namespace ExampleTOTP
{
    public enum TimeBasedPasswordStatus
    {
        NoUser,
        NoEmailAddress,
        NoSecretKey,
        Available
    }

    public abstract class TimeBasedPassword
    {
        // ------ Underlying user ID managed by derived implementation.
        protected object UserID { get; set; } = null;

        // ----- Methods that the derived class must provider.
        abstract protected void GetIDFromEmail();
        abstract protected string RetrieveSecretKey();
        abstract public bool HasSecretKey { get; }
        abstract protected bool StoreSecretKey(string secretKey);
        abstract protected bool IntervalAvailable(long whichInterval);
        abstract protected void DisableInterval(long whichInterval);

        private string EmailAddressInternal = null;
        public string EmailAddress
        {
            // ----- Email address for this instance's user.
            get
            {
                return this.EmailAddressInternal;
            }
            set
            {
                if (value == null)
                {
                    // ----- Clear ID as well.
                    this.EmailAddressInternal = null;
                    this.UserID = null;
                }
                else if (ValidateEmailAddress(value) == false)
					throw new ArgumentException($"Invalid email address '{value}'.",
						nameof(this.EmailAddress));
				else
                {
                    // ----- Give derived class a chance to locate the ID.
                    this.EmailAddressInternal = value;
                    this.GetIDFromEmail();
                }
            }
        }

        protected const int DefaultCodeLength = 6;
        private const int MinimumCodeLength = 6;
        private const int MaximumCodeLength = 8;
        private int CodeLengthInternal = DefaultCodeLength;
        public int CodeLength
        {
            // ----- The digit-length of the zero-padded code to be compared.
            get
            {
                return this.CodeLengthInternal;
            }
            set
            {
                if ((value < MinimumCodeLength) | (value > MaximumCodeLength))
                    throw new ArgumentOutOfRangeException(nameof(this.CodeLength),
                        $"Code length must range from {MinimumCodeLength} to {MaximumCodeLength}.");
				else
					this.CodeLengthInternal = value;
            }
        }

        const int DefaultKeyLength = 20; // 160 bits
        const int MinimumKeyLength = 16; // 128 bits
        const int MaximumKeyLength = 32; // 256 bits
        private int SecretKeyLengthInternal = DefaultKeyLength;
        public int SecretKeyLength
        {
            // ----- The length of the alphanumeric secret shared key.
            get
            {
                return this.SecretKeyLengthInternal;
            }
            set
            {
                if ((value < MinimumKeyLength) | (value > MaximumKeyLength))
                    throw new ArgumentOutOfRangeException(nameof(this.SecretKeyLength),
                        $"Secret key length must range from {MinimumKeyLength} to {MaximumKeyLength}.");
				else
					this.SecretKeyLengthInternal = value;
            }
        }

        const int DefaultIntervalLength = 30;
        const int MinimumIntervalLength = 5;
        const int MaximumIntervalLength = 600; // 10 minutes
        private int IntervalSecondsInternal = DefaultIntervalLength;
        public int IntervalSeconds
        {
            // ----- Size in seconds of each code-change interval.
            get
            {
                return this.IntervalSecondsInternal;
            }
            set
            {
                if ((value < MinimumIntervalLength) | (value > MaximumIntervalLength))
                    throw new ArgumentOutOfRangeException(nameof(this.IntervalSeconds),
                        $"Interval must range from {MinimumIntervalLength} to {MaximumIntervalLength}.");
				else
					this.IntervalSecondsInternal = value;
            }
        }

        const int DefaultDrift = 1;
        const int MinimumDrift = 0;
        const int MaximumDrift = 10;
        private int IntervalDriftInternal = DefaultDrift;
        public int IntervalDrift
        {
            // ----- Number of intervals before and after the target interval to check.
            get
            {
                return this.IntervalDriftInternal;
            }
            set
            {
                if ((value < MinimumDrift) | (value > MaximumDrift))
                    throw new ArgumentOutOfRangeException(nameof(this.IntervalDrift),
                        $"Interval drift must range from {MinimumDrift} to {MaximumDrift}.");
				else
					this.IntervalDriftInternal = value;
            }
        }

        // ----- Indicate the current state of the object.
        public TimeBasedPasswordStatus UseStatus
        {
            get
            {
                if (this.UserID == null)
                    return TimeBasedPasswordStatus.NoUser;
                else if (string.IsNullOrEmpty(this.EmailAddress))
                    return TimeBasedPasswordStatus.NoEmailAddress;
                else if (this.HasSecretKey == false)
                    return TimeBasedPasswordStatus.NoSecretKey;
                else
                    return TimeBasedPasswordStatus.Available;
            }
        }

        // ----- Display name for organization checking the code.
        public string Issuer { get; set; } = null;

        // ----- Internal access to the stored secret key for this user.
        //       Defers to the derived implementation.
        private string SecretKey
        {
            get
            {
                return this.RetrieveSecretKey();
            }
            set
            {
                if (this.StoreSecretKey(value) == false)
                    throw new Exception("Key could not be stored.");
            }
        }

        public bool GenerateKey(bool replaceExisting = false)
        {
            // ----- Create a new, full-random secret key for a user.
            byte[] randomBytes;
            char[] keyCharacters;

            // ----- The shared secret key can use any of these characters. The set is 64
            //       characters in total, which helps avoid selection bias.
            char[] allowedSet =
                ("ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
                "abcdefghijklmnopqrstuvwxyz" +
                "1234567890.-").ToCharArray();

            // ----- Don't replace an existing key unless forced to.
            if (this.HasSecretKey & !replaceExisting)
                return true;

            // ----- Generate a set of random values
            randomBytes = new byte[this.SecretKeyLength];
            using (RNGCryptoServiceProvider processing = new RNGCryptoServiceProvider())
                processing.GetBytes(randomBytes);

            // ----- Translate each unsigned byte value (0 to 255) to one of the
            //       characters (position 0 to 63).
            keyCharacters = new char[this.SecretKeyLength];
            for (int position = 0; position < this.SecretKeyLength; position++)
                keyCharacters[position] = allowedSet[randomBytes[position] % allowedSet.Length];

            // ----- Save the key to the user's record.
            try
            {
                this.SecretKey = new string(keyCharacters);
                return true;
            }
            catch
            {
                // ----- Treat all errors the same.
                return false;
            }
        }

        public bool DestroyKey()
        {
            // ----- Remove any defined key for this user. Return False on error.
            if (this.HasSecretKey)
                try
                {
                    this.SecretKey = null;
                }
                catch
                {
                    return false;
                }
            return true;
        }

        public string LinkPath
        {
            get
            {
                // ----- Return the sharing URI based on the othauth scheme.
                string linkBuilder;

                // ----- If something is wrong, return a blank URI.
                if (this.UseStatus != TimeBasedPasswordStatus.Available)
                    return "";

                // ----- Start with the scheme.
                linkBuilder = "otpauth://totp/";

                // ----- Add in the label, which is the user's email address, optionally paired
                //       with the name of the issuer.
                if (!string.IsNullOrWhiteSpace(this.Issuer))
                    linkBuilder += EscapeDataString(this.Issuer) + ":";
                linkBuilder += EscapeDataString(this.EmailAddress);

                // ----- Append the shared secret key, encoded as Base-32.
                linkBuilder += "?secret=" + Encode32(this.SecretKey);

                // ----- If the issuer exists, add it again.
                if (!string.IsNullOrWhiteSpace(this.Issuer))
                    linkBuilder += "&issuer=" + EscapeDataString(this.Issuer);

                // ----- Include any non-default interval size.
                if (this.IntervalSeconds != DefaultIntervalLength)
                    linkBuilder += $"&period={this.IntervalSeconds}";

                // ----- Include any non-default number of digits.
                if (this.CodeLength != DefaultCodeLength)
                    linkBuilder += $"&digits={this.CodeLength}";

                return linkBuilder;
            }
        }

        public bool CheckCode(string userCode, DateTime entryTime)
        {
            // ----- Check whether the numeric code entered by the user is correct for
            //       the indicated interval.
            long targetInterval;
            string officialCode;

            // ----- The user has to be sufficiently configured.
            if (this.UseStatus != TimeBasedPasswordStatus.Available)
                return false;

            // ----- Convert the date/time to an interval number since T0.
            targetInterval = IntervalFromTime(entryTime, this.IntervalSeconds);
            for (long testInterval = targetInterval - this.IntervalDrift;
                testInterval <= targetInterval + this.IntervalDrift; testInterval++)
            {
                // ---- Perhaps this interval was already used to grant access.
                if (this.IntervalAvailable(testInterval) == false)
                    continue;

                // ----- Test the code against the initial version.
                officialCode = this.GetCodeforInterval(this.SecretKey, testInterval);
                if (ConstantTimeCompare(officialCode, userCode))
                {
                    // ----- This code/interval can never be used again.
                    this.DisableInterval(testInterval);
                    return true;
                }
            }
            return false;
        }

        private string GetCodeforInterval(string secretKey, long timeIndex)
        {
            // ----- Generate the numeric code that will be compared to what the user enters.
            byte[] secretKeyBytes;
            HMACSHA1 hashProcessing;
            byte[] timeBits;
            byte[] fullHash;
            int startOffset;
            int theCode;

            // ----- This only works if everything is defined.
            if (this.UseStatus != TimeBasedPasswordStatus.Available)
                return "";

            // ----- The hash-generation tool expects the key to be stored as
            //       an array of bytes instead of as an intact string.
            secretKeyBytes = Encoding.ASCII.GetBytes(secretKey);

            // ----- The current interval (since the start of the Unix epoch) will be
            //       incorporated into the process to produce an interval-specific result.
            //       The algorithm requires "big endian" sorting.
            timeBits = BitConverter.GetBytes(timeIndex);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(timeBits);

            // ----- Use the HMAC-SHA-1 process to generate a hash.
            hashProcessing = new HMACSHA1(secretKeyBytes);
            fullHash = hashProcessing.ComputeHash(timeBits);

            // ----- The result will be many more (decimal) digits than are needed for
            //       the code comparison. The underlying HMAC-based one-time-password
            //       algorithm defines a truncation process.
            startOffset = fullHash[19] & 0xf;
            theCode = (fullHash[startOffset] & 0x7f) << 24
                    | (fullHash[startOffset + 1] & 0xff) << 16
                    | (fullHash[startOffset + 2] & 0xff) << 8
                    | (fullHash[startOffset + 3] & 0xff);

            // ----- Only take the least-significant digits, usually 6 digits total.
            theCode %= (int)Math.Pow(10, this.CodeLength);

            // ----- Zero-pad the result.
            return theCode.ToString($"D{this.CodeLength}");
        }
    }
}

